@@ -4,14 +4,12 @@
 
 
 # TODO:
-#   add tradingview realtime signals
-#   numpy
 #   scalper and dca bot
-#   unit-tests
 #   more docs and examples
 #   decimal
 #   3.9
 #   subplot plot directory
+#   decorator for strategies without exit condition (not converted data)
 
 from copy import copy
 from datetime import datetime
@@ -30,8 +28,8 @@ import ta.trend
 import ta.volatility
 import ta.volume
 from plotly.subplots import make_subplots
-from quick_trade import brokers
 from quick_trade import utils
+from quick_trade.brokers import TradingClient
 
 Line = dict  # To avoid the deprecation warning
 
@@ -67,7 +65,7 @@ class Trader(object):
     _backtest_out_no_drop: pd.DataFrame
     backtest_out: pd.DataFrame
     _open_lot_prices: List[float]
-    client: brokers.TradingClient
+    client: TradingClient
     __last_stop_loss: float
     __last_take_profit: float
     returns_strategy_diff: List[float]
@@ -85,11 +83,11 @@ class Trader(object):
                  interval: str = '1d',
                  trading_on_client: bool = True):
         ticker = ticker.upper()
-        assert isinstance(ticker, str), 'Deposit can only be a number.'
-        assert fullmatch(r'[A-Z]+/[A-Z]+', ticker), 'Ticker must match the pattern <[A-Z]+/[A-Z]+>'
-        assert isinstance(df, pd.DataFrame), 'Dataframe can only be of type pd.DataFrame.'
-        assert isinstance(interval, str), 'interval can only be of the string type.'
-        assert isinstance(trading_on_client, bool), 'trading_on_client can only be True or False.'
+        assert isinstance(ticker, str), 'The ticker can only be of type <str>.'
+        assert fullmatch(utils.TICKER_PATTERN, ticker), f'Ticker must match the pattern <{utils.TICKER_PATTERN}>'
+        assert isinstance(df, pd.DataFrame), 'Dataframe can only be of type <pd.DataFrame>.'
+        assert isinstance(interval, str), 'interval can only be of the <str> type.'
+        assert isinstance(trading_on_client, bool), 'trading_on_client can only be True or False (<bool>).'
 
         self.df = df.reset_index(drop=True)
         self.ticker = ticker
@@ -140,8 +138,6 @@ class Trader(object):
                 take = self._open_price
             if self._stop_loss is not np.inf:
                 _stop_loss = self._open_price
-        utils.logger.debug(
-            'stop loss: %f (%f pips), take profit: %f (%f pips)', _stop_loss, self._stop_loss, take, self._take_profit)
 
         return {'stop': _stop_loss,
                 'take': take}
@@ -156,7 +152,7 @@ class Trader(object):
         :return: (stop losses, take profits)
         """
         assert isinstance(add_stop_loss, (int, float)) and isinstance(add_take_profit, (int, float)), \
-            'Arguments to this function can only be float or int.'
+            'Arguments to this function can only be <float> or <int>.'
 
         utils.logger.debug('add stop-loss: %f pips, take-profit: %s pips', add_stop_loss, add_take_profit)
         stop_losses = []
@@ -582,7 +578,7 @@ class Trader(object):
         sell = buy.
         exit = exit.
         """
-        assert isinstance(swap_stop_take, (bool, int, float)), 'swap_stop_take can only be boolean or int(float)'
+        assert isinstance(swap_stop_take, (bool, int, float)), 'swap_stop_take can only be <bool>'
         if not isinstance(swap_stop_take, bool):
             msg = 'swap_stop_take is not a standard type'
             warn(msg)
@@ -621,12 +617,15 @@ class Trader(object):
         :param show: show the graph
         returns: pd.DataFrame with data of test
         """
-        assert isinstance(deposit, (float, int)), 'deposit must be of type int or float'
-        assert isinstance(bet, (float, int)), 'bet must be of type int or float'
-        assert isinstance(commission, (float, int)), 'commission must be of type int or float'
-        assert isinstance(plot, bool), 'plot must be of type bool'
-        assert isinstance(print_out, bool), 'print_out must be of type bool'
-        assert isinstance(show, bool), 'show must be of type bool'
+        assert isinstance(deposit, (float, int)), 'deposit must be of type <int> or <float>'
+        assert deposit > 0, 'deposit can\'t be 0 or less'
+        assert isinstance(bet, (float, int)), 'bet must be of type <int> or <float>'
+        assert bet > 0, 'bet can\'t be 0 or less'
+        assert isinstance(commission, (float, int)), 'commission must be of type <int> or <float>'
+        assert 0 <= commission < 100, 'commission cannot be >=100% or less then 0'
+        assert isinstance(plot, bool), 'plot must be of type <bool>'
+        assert isinstance(print_out, bool), 'print_out must be of type <bool>'
+        assert isinstance(show, bool), 'show must be of type <bool>'
 
         exit_take_stop: bool
         no_order: bool
@@ -692,20 +691,26 @@ class Trader(object):
                 ignore_breakout = True
 
             next_not_breakout = min(stop_loss, take_profit) < next_l <= next_h < max(stop_loss, take_profit)
-            if (min(stop_loss, take_profit) < low <= high < max(stop_loss,
-                                                                take_profit) and next_not_breakout) or ignore_breakout:
+
+            stop_loss = self._stop_losses[e - 1]
+            take_profit = self._take_profits[e - 1]
+            # be careful with e=0
+            now_not_breakout = min(stop_loss, take_profit) < low <= high < max(stop_loss,
+                                                                               take_profit)
+            if (ignore_breakout or now_not_breakout) and next_not_breakout:
                 diff = data_column[e + 1] - data_column[e]
             else:
                 exit_take_stop = True
-                if sig == utils.BUY and high >= take_profit:
-                    diff = take_profit - data_column[e]
 
-                elif sig == utils.BUY and low <= stop_loss:
+                if sig == utils.BUY and low <= stop_loss:
                     diff = stop_loss - data_column[e]
 
                 elif sig == utils.SELL and high >= stop_loss:
                     diff = stop_loss - data_column[e]
 
+                elif sig == utils.BUY and high >= take_profit:
+                    diff = take_profit - data_column[e]
+
                 elif sig == utils.SELL and low <= take_profit:
                     diff = take_profit - data_column[e]
 
@@ -740,6 +745,7 @@ class Trader(object):
             self.winrate = (self.profits / self.trades) * 100
         else:
             self.winrate = 0
+            utils.logger.error('0 trades in %s', self)
         self._info = f"""losses: {self.losses}
 trades: {self.trades}
 profits: {self.profits}
@@ -815,14 +821,16 @@ winrate: {self.winrate}%"""
                                                          'bprice': [],
                                                          'sprice': [],
                                                          'eprice': []}
-            for e, i in enumerate(converted):
-                if i == utils.SELL:
+            for e, (pred, conv, crlev) in enumerate(zip(self.returns,
+                                                        converted,
+                                                        utils.convert(self._credit_leverages))):
+                if conv == utils.SELL or (crlev is not np.nan and pred == utils.SELL):
                     preds['sellind'].append(e)
                     preds['sprice'].append(loc[e])
-                elif i == utils.BUY:
+                elif conv == utils.BUY or (crlev is not np.nan and pred == utils.BUY):
                     preds['buyind'].append(e)
                     preds['bprice'].append(loc[e])
-                elif i == utils.EXIT:
+                elif conv == utils.EXIT or (crlev is not np.nan and pred == utils.EXIT):
                     preds['exitind'].append(e)
                     preds['eprice'].append(loc[e])
             name: str
@@ -864,12 +872,24 @@ winrate: {self.winrate}%"""
                        plot: bool = True,
                        print_out: bool = True,
                        show: bool = True) -> pd.DataFrame:
-        assert isinstance(tickers, Iterable), 'tickers must be of type Iterable[str]'
+        assert isinstance(tickers, Iterable), 'tickers must be of type <Iterable[str]>'
         for el in tickers:
-            assert isinstance(el, str), 'tickers must be of type Iterable[str]'
-        assert isinstance(strategy_name, str), 'strategy_name must be of type str'
+            assert isinstance(el, str), 'tickers must be of type <Iterable[str]>'
+            assert fullmatch(utils.TICKER_PATTERN, el), f'all tickers must match the pattern <{utils.TICKER_PATTERN}>'
+        assert isinstance(strategy_name, str), 'strategy_name must be of type <str>'
         assert strategy_name in self.__dir__(), 'There is no such strategy'
         assert isinstance(strategy_kwargs, dict)
+        assert isinstance(limit, int), 'limit must be of type <int>'
+        assert limit > 0, 'limit can\'t be 0 or less'
+        assert isinstance(deposit, (float, int)), 'deposit must be of type <int> or <float>'
+        assert deposit > 0, 'deposit can\'t be 0 or less'
+        assert isinstance(bet, (float, int)), 'bet must be of type <int> or <float>'
+        assert bet > 0, 'bet can\'t be 0 or less'
+        assert isinstance(commission, (float, int)), 'commission must be of type <int> or <float>'
+        assert 0 <= commission < 100, 'commission cannot be >=100% or less then 0'
+        assert isinstance(plot, bool), 'plot must be of type <bool>'
+        assert isinstance(print_out, bool), 'print_out must be of type <bool>'
+        assert isinstance(show, bool), 'show must be of type <bool>'
 
         winrates: List[float] = []
         percentage_profits: List[float] = []
@@ -962,12 +982,12 @@ winrate: {self.winrate}%"""
         :param template: plotly template
         :param row_heights: standard
         """
-        assert isinstance(height, (int, float)), 'height must be of type int or float'
-        assert isinstance(width, (int, float)), 'width must be of type int or float'
-        assert isinstance(template, str), 'template must be of type str'
-        assert isinstance(row_heights, list), 'row_heights must be of type List[int, float]'
+        assert isinstance(height, (int, float)), 'height must be of type <int> or <float>'
+        assert isinstance(width, (int, float)), 'width must be of type <int> or <float>'
+        assert isinstance(template, str), 'template must be of type <str>'
+        assert isinstance(row_heights, list), 'row_heights must be of type <List[int, float]>'
         for el in row_heights:
-            assert isinstance(el, (int, float))
+            assert isinstance(el, (int, float)), 'row_heights must be of type <List[int, float]>'
 
         self.fig = make_subplots(3, 1, row_heights=row_heights, **subplot_kwargs)
         self.fig.update_layout(
@@ -985,6 +1005,7 @@ winrate: {self.winrate}%"""
             title_text=utils.DATA_TITLE, row=1, col=1, color=utils.TEXT_COLOR)
         utils.logger.info('new %s graph', self)
 
+    @utils.assert_logger
     def strategy_collider(self,
                           first_returns: utils.PREDICT_TYPE_LIST,
                           second_returns: utils.PREDICT_TYPE_LIST,
@@ -1052,6 +1073,8 @@ winrate: {self.winrate}%"""
 
         :return: combining of 2 strategies
         """
+        assert isinstance(first_returns, utils.PREDICT_TYPE_LIST) and isinstance(second_returns, utils.PREDICT_TYPE_LIST), \
+            'Arguments to this function can only be <utils.PREDICT_TYPE>.'
 
         if mode == 'minimalist':
             self.returns = []
@@ -1082,7 +1105,7 @@ winrate: {self.winrate}%"""
         return return_list
 
     @staticmethod
-    def _collide_super(l1, l2) -> utils.PREDICT_TYPE_LIST:
+    def _collide_super(l1: utils.PREDICT_TYPE_LIST, l2: utils.PREDICT_TYPE_LIST) -> utils.PREDICT_TYPE_LIST:
         return_list: utils.PREDICT_TYPE_LIST = []
         for first, sec in zip(utils.convert(l1), utils.convert(l2)):
             if first is not np.nan and sec is not np.nan and first is not sec:
@@ -1100,8 +1123,6 @@ winrate: {self.winrate}%"""
         if len(strategies) >= 3:
             for ret in strategies[2:]:
                 self.strategy_collider(self.returns, ret, mode=mode)
-        self.set_open_stop_and_take()
-        self.set_credit_leverages()
         return self.returns
 
     def get_trading_predict(self,
@@ -1147,8 +1168,8 @@ winrate: {self.winrate}%"""
                         bet = _moneys_
                     if coin_lotsize_division:
                         bet /= ticker_price
-                    self.client.exit_last_order()
 
+                    self.client.exit_last_order()
                     self.client.order_create(predict,
                                              self.ticker,
                                              bet * self.__last_credit_leverage)
@@ -1162,6 +1183,7 @@ winrate: {self.winrate}%"""
             'credit leverage': self.__last_credit_leverage
         }
 
+    @utils.assert_logger
     def realtime_trading(self,
                          strategy,
                          ticker: str = 'BTC/USDT',
@@ -1189,6 +1211,18 @@ winrate: {self.winrate}%"""
         :param strategy_kwargs: named arguments to -strategy.
         :param strategy_args: arguments to -strategy.
         """
+        assert fullmatch(utils.TICKER_PATTERN, ticker), f'Ticker must match the pattern <{utils.TICKER_PATTERN}>'
+        assert isinstance(print_out, bool), 'print_out must be of type <bool>'
+        assert isinstance(bet_for_trading_on_client,
+                          (float, int)), 'bet_for_trading_on_client must be of type <float> or <int>'
+        assert isinstance(ignore_exceptions, bool), 'ignore_exceptions must be of type <bool>'
+        assert isinstance(print_exc, bool), 'print_exc must be of type <bool>'
+        assert isinstance(wait_sl_tp_checking, (float, int)), 'wait_sl_tp_checking must be of type <float> or <int>'
+        assert wait_sl_tp_checking < self._sec_interval, \
+            'wait_sl_tp_checking cannot be greater than or equal to the timeframe'
+        assert isinstance(limit, int), 'limit must be of type <int>'
+        assert isinstance(strategy_in_sleep, bool), 'strategy_in_sleep must be of type <bool>'
+        assert isinstance(coin_lotsize_division, bool), 'coin_lotsize_division must be of type <bool>'
 
         self.ticker = ticker
         open_time = time()
@@ -1243,6 +1277,7 @@ winrate: {self.winrate}%"""
                 else:
                     raise exc
 
+    @utils.assert_logger
     def multi_realtime_trading(self,
                                tickers: List[str],
                                start_time: datetime,  # LOCAL TIME
@@ -1257,6 +1292,28 @@ winrate: {self.winrate}%"""
                                strategy_in_sleep: bool = False,
                                deposit_part: Union[float, int] = 1.0,  # for all trades
                                **strategy_kwargs):
+        assert isinstance(tickers, Iterable), 'tickers must be of type <Iterable[str]>'
+        for el in tickers:
+            assert isinstance(el, str), 'tickers must be of type <Iterable[str]>'
+            assert fullmatch(utils.TICKER_PATTERN, el), f'all tickers must match the pattern <{utils.TICKER_PATTERN}>'
+        assert isinstance(print_out, bool), 'print_out must be of type <bool>'
+        assert isinstance(bet_for_trading_on_client,
+                          (float, int)), 'bet_for_trading_on_client must be of type <float> or <int>'
+        assert isinstance(ignore_exceptions, bool), 'ignore_exceptions must be of type <bool>'
+        assert isinstance(print_exc, bool), 'print_exc must be of type <bool>'
+        assert isinstance(wait_sl_tp_checking, (float, int)), 'wait_sl_tp_checking must be of type <float> or <int>'
+        assert wait_sl_tp_checking < self._sec_interval, \
+            'wait_sl_tp_checking cannot be greater than or equal to the timeframe'
+        assert isinstance(limit, int), 'limit must be of type <int>'
+        assert isinstance(strategy_in_sleep, bool), 'strategy_in_sleep must be of type <bool>'
+        assert isinstance(start_time, datetime), 'start_time must be of type <datetime.datetime>'
+        assert start_time > datetime.now(), 'start_time cannot be earlier than the present time'
+        assert isinstance(strategy_name, str), 'strategy_name must be of type <str>'
+        assert strategy_name in self.__dir__(), 'There is no such strategy'
+        assert isinstance(coin_lotsize_division, bool), 'coin_lotsize_division must be of type <bool>'
+        assert isinstance(deposit_part, (int, float)), 'deposit_part must be of type <int> or <float>'
+        assert 1 >= deposit_part > 0, 'deposit_part cannot be greater than 1 or less than 0(inclusively)'
+
         can_orders: int = len(tickers)
         bet_for_trading_on_client_copy: Union[float, int] = bet_for_trading_on_client
 
@@ -1266,7 +1323,7 @@ winrate: {self.winrate}%"""
                                     coin_lotsize_division: bool = True
                                     ) -> Dict[str, Union[str, float]]:
                 balance = self.client.get_balance_ticker(self.ticker.split('/')[1])
-                bet = (balance * 10) / (can_orders / deposit_part - brokers.TradingClient.cls_open_orders)
+                bet = (balance * 10) / (can_orders / deposit_part - TradingClient.cls_open_orders)
                 bet /= 10  # decimal analog
                 if bet > bet_for_trading_on_client_copy:
                     bet = bet_for_trading_on_client_copy
@@ -1310,16 +1367,23 @@ winrate: {self.winrate}%"""
         self.fig.update_yaxes(row=3, col=1, type='log')
         utils.logger.debug('trader log returns')
 
-    def set_client(self, your_client: brokers.TradingClient):
+    @utils.assert_logger
+    def set_client(self, your_client: TradingClient):
         """
         :param your_client: trading client
         """
+        assert isinstance(your_client, TradingClient), 'your_client must be of type <TradingClient>'
+
         self.client = your_client
         utils.logger.debug('trader set client')
 
+    @utils.assert_logger
     def convert_signal(self,
                        old: utils.PREDICT_TYPE = utils.SELL,
                        new: utils.PREDICT_TYPE = utils.EXIT) -> utils.PREDICT_TYPE_LIST:
+        assert isinstance(old, utils.PREDICT_TYPE) and isinstance(new, utils.PREDICT_TYPE), \
+            'Arguments to this function can only be <utils.PREDICT_TYPE>.'
+
         pos: int
         val: utils.PREDICT_TYPE
         for pos, val in enumerate(self.returns):
@@ -1328,6 +1392,7 @@ winrate: {self.winrate}%"""
         utils.logger.debug("trader signals converted: %s >> %s", old, new)
         return self.returns
 
+    @utils.assert_logger
     def set_open_stop_and_take(self,
                                take_profit: Union[float, int] = np.inf,
                                stop_loss: Union[float, int] = np.inf,
@@ -1339,6 +1404,8 @@ winrate: {self.winrate}%"""
         :param take_profit: take profit in points
         :param stop_loss: stop loss in points
         """
+        # TODO
+
         self._take_profit = take_profit
         self._stop_loss = stop_loss
         take_flag: float = np.inf
